package game;

import sk.thenet.plustd.FM;
import sk.thenet.gamestd.bmp.*;

enum Layer {
	Static(bd:BitmapData);
	Entities;
	Parallax(bd:BitmapData, fx:Float, fy:Float);
	Tree;
	Water(w:Water);
}

class Level {
	public static var LEVELS:Array<Level>;
	
	public static function initLevels():Void {
		LEVELS = [
			new Level("hub", 0, 0, 20, 22, function(lvl:Level, phsx:Physics){
				new EManInactive(16 * 5 + 8, 16 * 18 + 8, lvl, phsx, 0);
				new EManInactive(16 * 6 + 8, 16 * 18 + 8, lvl, phsx, 1);
				new EManInactive(16 * 7 + 8, 16 * 18 + 8, lvl, phsx, 2);
				lvl.man = new EMan(16 * 9, 16 * 11 + 8, lvl, phsx);
				lvl.camxmin = 0;
				lvl.camxmax = 0;
				lvl.camymin = 0;
				lvl.camymax = 80;
				lvl.camspeed = 53;
			}, 0xFF1C1B1C, [
				Static(Assets.aLvlHub),
				Entities,
				Tree
			]),
			new Level("grotto", 24, 4, 33, 37, function(lvl:Level, phsx:Physics){
				lvl.man = new EMan(16 * 27, 16 * 21, lvl, phsx);
				new EESnail(16 * 24, 16 * 21, lvl, phsx);
				new EESnail(16 * 1, 16 * 12, lvl, phsx);
				new EESnail(16 * 2, 16 * 11, lvl, phsx);
				new EESnail(16 * 3, 16 * 10, lvl, phsx);
				new EESnail(16 * 2, 16 * 9, lvl, phsx);
				new EECrab(16 * 29 + 8, 16 * 10, lvl, phsx);
				new EECrab(16 * 18 + 8, 16 * 2, lvl, phsx);
				lvl.camxmin = 0;
				lvl.camxmax = 13 * 16;
				lvl.camymin = 0;
				lvl.camymax = 22 * 16;
				lvl.camspeed = 53;
			}, 0xFF255920, [
				Parallax(Assets.aLvlGrotto.cut(528 * 3, 0, 528, 592), .1, .1),
				Parallax(Assets.aLvlGrotto.cut(528 * 2, 0, 528, 592), .2, .2),
				Parallax(Assets.aLvlGrotto.cut(528, 0, 528, 592), .4, .4),
				Static(Assets.aLvlGrotto.cut(0, 0, 528, 592)),
				Entities,
				Water(new Water(15 * 16, 24 * 16, 6 * 16, 8 * 16, null)),
				Water(new Water(8 * 16, 34 * 16, 4 * 16, 1 * 16, null)),
				Static(Assets.aLvlGrotto.cut(528 * 4, 0, 528, 592))
			]),
			new Level("shrooms", 0, 43, 62, 17, function(lvl:Level, phsx:Physics){
				lvl.man = new EMan(16 * 40, 16 * 5, lvl, phsx);
				new EEShrool(16 * 40, 16 * 9, lvl, phsx);
				new EEShrool(16 * 30, 16 * 9, lvl, phsx);
				new EEShrool(16 * 11, 16 * 4, lvl, phsx);
				new EEBoss(16 * 57, 16 * 12, lvl, phsx);
				lvl.camxmin = 0;
				lvl.camxmax = 42 * 16;
				lvl.camymin = 0;
				lvl.camymax = 2 * 16;
				lvl.camspeed = 53;
			}, Assets.palette[17], [
				Parallax(Assets.aLvlShrooms.cut(0, 272, 992, 272), .4, 0),
				Static(Assets.aLvlShrooms.cut(0, 0, 992, 272)),
				Entities,
				Water(new Water(2 * 16, 12 * 16, 6 * 16, 2 * 16, null)),
				Water(new Water(19 * 16, 14 * 16, 8 * 16, 2 * 16, null)),
			])
		];
	}
	
	public static function resetLevels():Void {
		for (i in 0...LEVELS.length){
			for (j in 0...LEVELS[i].state.length){
				LEVELS[i].state[j] = true;
			}
		}
	}
	
	public static function getLevel(id:String):Level {
		for (l in LEVELS){
			if (l.id == id) return l;
		}
		return null;
	}
	
	public var id:String;
	public var width:Int;
	public var height:Int;
	public var lay:Array<Layer>;
	public var bgcol:UInt;
	public var colmap:Array<Int>;
	public var ent:Array<Entity>;
	public var init_ent:Array<Level->Physics->Void>;
	public var man:EMan;
	public var init:Level->Physics->Void;
	public var camx:Float;
	public var camy:Float;
	public var camxmin:Int;
	public var camxmax:Int;
	public var camymin:Int;
	public var camymax:Int;
	public var camspeed:Int;
	public var camxi:Int;
	public var camyi:Int;
	public var state:Array<Bool>;
	
	public function new(nid:String, ncolx:Int, ncoly:Int, ncolw:Int, ncolh:Int, ninit:Level->Physics->Void, nbgcol:UInt, nlay:Array<Layer>){
		var ncolmap = Assets.aLevels.cut(ncolx, ncoly, ncolw, ncolh);
		var ngravmap = Assets.aLevelsGravity.cut(ncolx, ncoly, ncolw, ncolh);
		var nonemap = Assets.aLevelsOneway.cut(ncolx, ncoly, ncolw, ncolh);
		var npickmap = Assets.aLevelsPickups.cut(ncolx, ncoly, ncolw, ncolh);
		ent = [];
		init_ent = [];
		id = nid;
		width = ncolw;
		height = ncolh;
		init = ninit;
		bgcol = nbgcol;
		lay = nlay;
		colmap = [];
		var i:Int = 0;
		var si:Int = 0;
		for (y in 0...height){
			for (x in 0...width){
				colmap[i] = (switch (ncolmap.get(x, y) & 0xFFFFFF){
					case 0: Physics.P_LVL_FULL;
					case 0xFF0000: Physics.P_LVL_SLOPE_11_NW;
					case 0xFFFF00: Physics.P_LVL_SLOPE_11_NE;
					case _: Physics.P_LVL_EMPTY;
				}) | (switch (ngravmap.get(x, y) & 0xFFFFFF){
					case 0x0000FF: Physics.P_LVL_WATER;
					case 0x00FFFF: Physics.P_LVL_LADDER;
					case _: Physics.P_LVL_EMPTY;
				}) | (switch (nonemap.get(x, y) & 0xFFFFFF){
					case 0xFF0000: Physics.P_LVL_NO_LEFT | Physics.P_LVL_NO_RIGHT | Physics.P_LVL_NO_BOTTOM | Physics.P_LVL_FULL;
					case 0x00FF00: Physics.P_LVL_NO_LEFT | Physics.P_LVL_NO_BOTTOM | Physics.P_LVL_NO_TOP | Physics.P_LVL_FULL;
					case 0xFF00FF: Physics.P_LVL_NO_RIGHT | Physics.P_LVL_NO_BOTTOM | Physics.P_LVL_FULL;
					case 0x0000FF: Physics.P_LVL_NO_RIGHT | Physics.P_LVL_NO_BOTTOM | Physics.P_LVL_NO_TOP | Physics.P_LVL_FULL;
					case _: Physics.P_LVL_EMPTY;
				});
				switch (npickmap.get(x, y) & 0xFFFFFF){
					case 0xFF0000: init_ent.push(function(lvl:Level, phsx:Physics, csi:Int){
						if (lvl.state[csi])
							new EPickup(16 * x, 16 * y, lvl, phsx, true, false, csi);
					}.bind(_, _, si++));
					case 0xFFFF00: init_ent.push(function(lvl:Level, phsx:Physics, csi:Int){
						if (lvl.state[csi])
							new EPickup(16 * x, 16 * y, lvl, phsx, false, false, csi);
					}.bind(_, _, si++));
					case 0x00FF00: init_ent.push(function(lvl:Level, phsx:Physics, csi:Int){
						if (lvl.state[csi])
							new EPickup(16 * x, 16 * y, lvl, phsx, true, true, csi);
					}.bind(_, _, si++));
					case 0x00FFFF: init_ent.push(function(lvl:Level, phsx:Physics, csi:Int){
						if (lvl.state[csi])
							new EPickup(16 * x, 16 * y, lvl, phsx, false, true, csi);
					}.bind(_, _, si++));
				}
				i++;
			}
		}
		state = [
			for (i in 0...si){
				true;
			}
		];
	}
	
	public function tick():Void {
		ent = ent.filter(function(a:Entity){
			return !a.remove;
		});
		for (e in ent){
			e.tick();
		}
	}
	
	public function kUp(k:Int):Void {
		for (e in ent){
			e.kUp(k);
		}
	}
	
	public function kDown(k:Int):Void {
		for (e in ent){
			e.kDown(k);
		}
	}
	
	public function render(bd:BitmapData):Void {
		bd.fill(bgcol);
		camx = (camx * camspeed + man.pe.xp - 160) / (camspeed + 1);
		camy = (camy * camspeed + man.pe.yp - 120) / (camspeed + 1);
		camxi = FM.floor(camx);
		camyi = FM.floor(camy);
		if (camxi < camxmin) camxi = camxmin;
		if (camxi > camxmax) camxi = camxmax;
		if (camyi < camymin) camyi = camymin;
		if (camyi > camymax) camyi = camymax;
		for (l in lay){
			switch (l){
				case Static(lbd):
				bd.copyPixels(lbd, -camxi, -camyi, 0, 0, lbd.width, lbd.height);
				case Entities:
				for (e in ent){
					if (!e.invisible) e.render(bd, camxi, camyi);
				}
				case Parallax(lbd, fx, fy):
				var camdx:Int = (lbd.width >> 1) - camxi;
				var camdy:Int = (lbd.height >> 1) - camyi;
				bd.copyPixels(lbd, -(lbd.width >> 1) + 160 + FM.floor(camdx * fx),
					-(lbd.height >> 1) + 120 + FM.floor(camdy * fy), 0, 0, lbd.width, lbd.height);
				case Water(w):
				w.render(bd, camxi, camyi);
				case Tree:
				cast(Main.st.cstr, GSGame).tree.render(bd, camxi, camyi);
			}
		}
	}
}